import os
import z3c.relationfield
from zope.app.testing.functional import ZCMLLayer

ftesting_zcml = os.path.join(
    os.path.dirname(z3c.relationfield.__file__), 'ftesting.zcml')
FunctionalLayer = ZCMLLayer(ftesting_zcml, __name__, 'FunctionalLayer')



from z3c.relationfield.relation import (RelationValue,
                                        TemporaryRelationValue,
                                        create_relation)
from z3c.relationfield.index import RelationCatalog
from z3c.relationfield.schema import Relation, RelationList
from z3c.relationfield.event import realize_relations



import grokcore.component as grok

from persistent import Persistent
from zope.interface import implements, providedBy, Declaration
from zope import component
from zope.app.intid.interfaces import IIntIds

from z3c.objpath.interfaces import IObjectPath

from z3c.relationfield.interfaces import (IRelationValue,
                                          ITemporaryRelationValue)

class RelationValue(Persistent):
    implements(IRelationValue)

    def __init__(self, to_id):
        self.to_id = to_id
        # these will be set automatically by RelationProperty
        self.from_object = None
        self.__parent__ = None
        self.from_attribute = None

    @property
    def from_id(self):
        intids = component.getUtility(IIntIds)
        return intids.getId(self.from_object)

    @property
    def from_path(self):
        return _path(self.from_object)

    @property
    def from_interfaces(self):
        return providedBy(self.from_object)

    @property
    def from_interfaces_flattened(self):
        return _interfaces_flattened(self.from_interfaces)
        
    @property
    def to_object(self):
        return _object(self.to_id)

    @property
    def to_path(self):
        return _path(self.to_object)

    @property
    def to_interfaces(self):
        return providedBy(self.to_object)

    @property
    def to_interfaces_flattened(self):
        return _interfaces_flattened(self.to_interfaces)

    def __cmp__(self, other):
        if other is None:
            return cmp(self.to_id, None)
        return cmp(self.to_id, other.to_id)

class TemporaryRelationValue(Persistent):
    """A relation that isn't fully formed yet.

    It needs to be finalized afterwards, when we are sure all potential
    target objects exist.
    """
    grok.implements(ITemporaryRelationValue)
    
    def __init__(self, to_path):
        self.to_path = to_path

    def convert(self):
        object_path = component.getUtility(IObjectPath)
        # XXX what if we have a broken relation?
        to_object = object_path.resolve(self.to_path)
        intids = component.getUtility(IIntIds)
        to_id = intids.getId(to_object)
        return RelationValue(to_id)
    
def _object(id):
    intids = component.getUtility(IIntIds)
    try:
        return intids.getObject(id)
    except KeyError:
        # XXX catching this error is not the right thing to do.
        # instead, breaking a relation by removing an object should
        # be caught and the relation should be adjusted that way.
        return None

def _path(obj):
    if obj is None:
        return ''
    object_path = component.getUtility(IObjectPath)
    return object_path.path(obj)

def _interfaces_flattened(interfaces):
    return Declaration(*interfaces).flattened()

def create_relation(obj):
    intids = component.getUtility(IIntIds)
    return RelationValue(intids.getId(obj))



import unittest

from zope.app.testing.setup import setUpTestAsModule, tearDownTestAsModule
from zope.app.testing.functional import FunctionalDocFileSuite
from z3c.relationfield.testing import FunctionalLayer

def setUp(test):
    setUpTestAsModule(test, name='__builtin__')

def test_suite():
    globs = {}
    readme = FunctionalDocFileSuite(
        'README.txt',
        globs = globs,
        setUp=setUp,
        tearDown=tearDownTestAsModule,
        )
    readme.layer = FunctionalLayer
    return unittest.TestSuite([readme])



from setuptools import setup, find_packages
import sys, os

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description = (
    read('src', 'z3c', 'relationfield', 'README.txt')
    + '\n' +
    read('CHANGES.txt')
    + '\n' +
    'Download\n'
    '********\n'
    )

setup(
    name='z3c.relationfield',
    version='0.2',
    description="A relation field framework for Zope 3.",
    long_description=long_description,
    classifiers=[],
    keywords='',
    author='Martijn Faassen',
    author_email='faassen@startifact.com',
    license='',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['z3c'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'ZODB3',
        'grokcore.component',
        'zope.app.intid',
        'zope.app.container',
        'z3c.schema2xml >= 1.0',
        'z3c.objpath',
        'zc.relation >= 1.0',
        # for tests
        'zope.app.zcmlfiles',
        'zope.securitypolicy',
        ],
    entry_points={},
    )



from zope.interface import Interface, Attribute
from zope.schema.interfaces import IField, IList

class IHasRelations(Interface):
    """Marker interface indicating that the object has relations.

    Use this interface to make sure that the relations get added and
    removed from the catalog when appropriate.
    """

class IRelation(IField):
    pass

class IRelationList(IList):
    pass

class IRelationValue(Interface):
    """A relation between the parent object and another one.

    This should be stored as the value in the object when the schema uses the
    Relation field.
    """
    from_object = Attribute("The object this relation is pointing from.")

    from_id = Attribute("Id of the object this relation is pointing from.")
    
    from_path = Attribute("The path of the from object.")

    from_interfaces = Attribute("The interfaces of the from object.")

    from_interfaces_flattened = Attribute(
        "Interfaces of the from object, flattened. "
        "This includes all base interfaces.")
    
    from_attribute = Attribute("The name of the attribute of the from object.")
    
    to_object = Attribute("The object this relation is pointing to.")

    to_id = Attribute("Id of the object this relation is pointing to.")

    to_path = Attribute("The path of the object this relation is pointing to.")

    to_interfaces = Attribute("The interfaces of the to-object.")

    to_interfaces_flattened = Attribute(
        "The interfaces of the to object, flattened. "
        "This includes all base interfaces.")

class ITemporaryRelationValue(Interface):
    """A temporary relation.

    When importing relations from XML, we cannot resolve them into
    true RelationValue objects yet, as it may be that the object that is
    being related to has not yet been loaded. Instead we create
    a TemporaryRelationValue object that can be converted into a real one
    after the import has been concluded.
    """
    def convert():
        """Convert temporary relation into a real one.

        Returns real relation object
        """



import grokcore.component as grok

from zope.interface import providedBy
from zope.schema import getFields
from zope import component
from zope.app.intid.interfaces import IIntIds
from zope.app.container.interfaces import (IObjectAddedEvent,
                                           IObjectRemovedEvent)
from zope.lifecycleevent.interfaces import IObjectModifiedEvent

from zc.relation.interfaces import ICatalog

from z3c.relationfield.interfaces import (IHasRelations,
                                          IRelation, IRelationList,
                                          IRelationValue,
                                          ITemporaryRelationValue)

@grok.subscribe(IHasRelations, IObjectAddedEvent)
def addRelations(obj, event):
    """Register relations.

    Any relation object on the object will be added.
    """
    for name, relation in _relations(obj):
         _setRelation(obj, name, relation)

@grok.subscribe(IHasRelations, IObjectRemovedEvent)
def removeRelations(obj, event):
    """Remove relations.

    Any relation object on the object will be removed from the catalog.
    """
    catalog = component.getUtility(ICatalog)
 
    for name, relation in _relations(obj):
        if relation is not None:
            catalog.unindex(relation)

@grok.subscribe(IHasRelations, IObjectModifiedEvent)
def updateRelations(obj, event):
    """Re-register relations, after they have been changed.
    """
    catalog = component.getUtility(ICatalog)
    intids = component.getUtility(IIntIds)

    # remove previous relations coming from id (now have been overwritten)
    # have to activate query here with list() before unindexing them so we don't
    # get errors involving buckets changing size
    rels = list(catalog.findRelations({'from_id': intids.getId(obj)}))
    for rel in rels:
        catalog.unindex(rel)    

    # add new relations
    addRelations(obj, event)

def realize_relations(obj):
    """Given an object, convert any temporary relations on it to real ones.
    """
    for name, index, relation in _potential_relations(obj):
        if ITemporaryRelationValue.providedBy(relation):
            if index is None:
                # relation
                setattr(obj, name, relation.convert())
            else:
                # relation list
                getattr(obj, name)[index] = relation.convert()

def _setRelation(obj, name, value):
    """Set a relation on an object.

    Sets up various essential attributes on the relation.
    """
    # if the Relation is None, we're done
    if value is None:
        return
    # make sure relation has a __parent__ so we can make an intid for it
    value.__parent__ = obj
    # also set from_object to parent object
    value.from_object = obj
    # and the attribute to the attribute name
    value.from_attribute = name
    # now we can create an intid for the relation
    intids = component.getUtility(IIntIds)
    id = intids.register(value)
    # and index the relation with the catalog
    catalog = component.getUtility(ICatalog)
    catalog.index_doc(id, value)

def _relations(obj):
    """Given an object, return tuples of name, relation value.

    Only real relations are returned, not temporary relations.
    """
    for name, index, relation in _potential_relations(obj):
        if IRelationValue.providedBy(relation):
            yield name, relation

def _potential_relations(obj):
    """Given an object return tuples of name, index, relation value.

    Returns both IRelationValue attributes as well as ITemporaryRelationValue
    attributes.

    If this is a IRelationList attribute, index will contain the index
    in the list. If it's a IRelation attribute, index will be None.
    """
    for iface in providedBy(obj).flattened():
        for name, field in getFields(iface).items():
            if IRelation.providedBy(field):
                try:
                    relation = getattr(obj, name)
                except AttributeError:
                    # can't find this relation on the object
                    continue
                yield name, None, relation
            if IRelationList.providedBy(field):
                try:
                    l = getattr(obj, name)
                except AttributeError:
                    # can't find the relation list on this object
                    continue
                if l is not None:
                    for i, relation in enumerate(l):
                        yield name, i, relation



import grokcore.component as grok

from lxml import etree

from zope.interface import implements
from zope.schema import Field, List

from z3c.schema2xml import IXMLGenerator

from z3c.relationfield.interfaces import IRelation, IRelationList
from z3c.relationfield.relation import TemporaryRelationValue

class Relation(Field):
    implements(IRelation)

class RelationGenerator(grok.Adapter):
    """Eport a relation to XML.
    """
    grok.context(IRelation)
    grok.implements(IXMLGenerator)

    def output(self, container, value):
        element = etree.SubElement(container, self.context.__name__)
        if value is not None:
            element.text = value.to_path

    def input(self, element):
        if element.text is None:
            return None
        path = element.text
        return TemporaryRelationValue(path)

class RelationList(List):
    implements(IRelationList)

    value_type = Relation()
    
class RelationListGenerator(grok.Adapter):
    """Export a relation list to XML.
    """
    grok.context(IRelationList)
    grok.implements(IXMLGenerator)

    def output(self, container, value):
        element = etree.SubElement(container, self.context.__name__)
        field = self.context.value_type
        if value is not None:
            for v in value:
                IXMLGenerator(field).output(element, v)

    def input(self, element):
        field = self.context.value_type
        return [
            IXMLGenerator(field).input(sub_element)
            for sub_element in element]




import BTrees

from zope import component
from zope.app.intid.interfaces import IIntIds

from zc.relation.catalog import Catalog
from zc.relation.interfaces import ICatalog

from z3c.relationfield.interfaces import IRelationValue

def dump(obj, catalog, cache):
    intids = cache.get('intids')
    if intids is None:
        intids = cache['intids'] = component.getUtility(IIntIds)
    return intids.getId(obj)
    
def load(token, catalog, cache):
    intids = cache.get('intids')
    if intids is None:
        intids = cache['intids'] = component.getUtility(IIntIds)
    return intids.getObject(token)
    
class RelationCatalog(Catalog):

    def __init__(self):
        Catalog.__init__(self, dump, load)
        self.addValueIndex(IRelationValue['from_id'])
        self.addValueIndex(IRelationValue['to_id'])
        self.addValueIndex(IRelationValue['from_attribute'],
                           btree=BTrees.family32.OI)
        self.addValueIndex(IRelationValue['from_interfaces_flattened'],
                           multiple=True,
                           btree=BTrees.family32.OI)
        self.addValueIndex(IRelationValue['to_interfaces_flattened'],
                           multiple=True,
                           btree=BTrees.family32.OI)

